
/******************************************************************************
* MODULE     : fft_root.hpp
* DESCRIPTION: arithmetic on 2^k-th roots of unity used by the FFT
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__FFT_ROOT__HPP
#define __MMX__FFT_ROOT__HPP
#include "modular.hpp"

namespace mmx {
#define TMPL template<class C>
#define Fft_root fft_root<C>

/******************************************************************************
* Primitive roots of unity
******************************************************************************/

TMPL
class fft_root {
MMX_ALLOCATORS;
public:
  C u;
  inline fft_root (): u (1) {}
  fft_root (nat n);
  inline fft_root (const C& c): u (c) {}
  inline Fft_root& operator = (const Fft_root& x) {
    u= x.u; return *this; }
  inline friend bool hard_eq (const Fft_root& u, const Fft_root& v) {
    return hard_eq (u.u, v.u); }
  inline friend bool hard_neq (const Fft_root& u, const Fft_root& v) {
    return hard_neq (u.u, v.u); }
  inline friend bool eq (const Fft_root& u, const Fft_root& v) {
    return eq (u.u, v.u); }
  inline friend bool neq (const Fft_root& u, const Fft_root& v) {
    return neq (u.u, v.u); }
  inline bool operator == (const Fft_root& v) { return u == v.u; }
  inline bool operator != (const Fft_root& v) { return u != v.u; }
};

TMPL C inline
operator * (const C& c, const Fft_root& u) {
  return c * u.u;
}

TMPL Fft_root inline
operator * (const Fft_root& u, const Fft_root& v) {
  return Fft_root (u.u * v.u);
}

TMPL inline generic flatten (const Fft_root& u) {
  return flatten (u.u); }

PRINT_SUGAR(TMPL,Fft_root)

/******************************************************************************
* Special operations for FFT
******************************************************************************/

#define FFT_DONT_COUNT
#ifdef FFT_COUNT
static nat cross_count;
static nat icross_count;
static nat flip_count;
#endif

TMPL inline void
fft_cross (C* c1, C* c2, const Fft_root* u) {
  C temp= (*c2) * (*u);
  *c2   = (*c1) - temp;
  *c1   = (*c1) + temp;
#ifdef FFT_COUNT
  cross_count++;
#endif
}

TMPL inline void
fft_cross (C* c1, C* c2) {
  C temp= (*c2);
  *c2   = (*c1) - temp;
  *c1   = (*c1) + temp;
#ifdef FFT_COUNT
  cross_count++;
#endif
}

TMPL inline void
ifft_cross (C* c1, C* c2, const Fft_root* iu) {
  C temp= *c2;
  *c2   = ((*c1) - temp) * (*iu);
  *c1   = (*c1) + temp;
#ifdef FFT_COUNT
  icross_count++;
#endif
}

TMPL inline void
ifft_cross (C* c1, C* c2) {
  C temp= *c2;
  *c2   = (*c1) - temp;
  *c1   = (*c1) + temp;
#ifdef FFT_COUNT
  icross_count++;
#endif
}

TMPL inline void
itft_flip (C* c1, C* c2, const Fft_root* u) {
  C temp= (*c2) * (*u);
  *c1   = (*c1) - temp;
  *c2   = (*c1) - temp;
#ifdef FFT_COUNT
  flip_count++;
#endif
}

TMPL inline void
itft_flip (C* c1, C* c2) {
  *c1   = (*c1) - (*c2);
  *c2   = (*c1) - (*c2);
#ifdef FFT_COUNT
  flip_count++;
#endif
}

TMPL inline void
fft_shift (C* dest, nat n) {
  nat i= n;
  C u= invert (C((int) n));
  while (i!=0) {
    *dest= u * (*dest);
    dest++; i--;
  }
}

/******************************************************************************
* A particularly nice modulus for FFT: p= 3*2^30+1
******************************************************************************/

#define FFTP nat_value<0x0c0000001>

template<>
fft_root<modular<FFTP> >::fft_root (nat n) {
  modular<FFTP> v= 125;
  assert ((1<<30)%n == 0); // |n| must be a power of two
  u= binpow (v, (1<<30) / n);
}

#undef TMPL
#undef Fft_root
} // namespace mmx
#endif //__MMX__FFT_ROOT__HPP
