
/******************************************************************************
* MODULE     : skew_jet.hpp
* DESCRIPTION: low level arithmetic for skew polynomials
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__SKEW_JET__HPP
#define __MMX__SKEW_JET__HPP
#include "jet.hpp"
#include "karatsuba.hpp"

namespace mmx {

/******************************************************************************
* Skew multiplication variant
******************************************************************************/

template<typename Op> class skew_multiply {};

/******************************************************************************
* Prototypes
******************************************************************************/

template<typename C, typename Op>
class adjoint_helper {
public:
  static void indeterminate (C* dest) {
    // assigns dest (of length 2) with adjoint of the unit
    assert (false); // adjoint of indeterminate not implemented for this Op
  }
  static void mul (C* dest, const C* s1, const C* s2,
		   nat n1, nat n2, bool adj) {
    // multiplication for the adjoint operator of Op if adj holds
    assert (false); // adjoint multiplication not implemented for this Op
  }
};

/******************************************************************************
* The operator d/dx
******************************************************************************/

class d_operator {};

template<typename C> inline void
jet_coeff_derive (C* dest, nat n) {
  while (n != 0) {
    *dest= derive (*dest);
    dest++; n--;
  }
}

template<typename C>
class multiplier<C,skew_multiply<d_operator>,C > {
public:
  static void mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    if (n1 == 0 || n2 == 0) return;
    INLINE_ARRAY (C, buf, (n1+n2)<<1);
    C* t1 = buf;
    C* t2 = t1 + n1;
    C* aux= t2 + n2;
    jet_clear (dest, n1+n2-1);
    jet_copy (t1, s1, n1);
    jet_copy (t2, s2, n2);
    int i=0;
    C p (1);
    while (n1 > 0) {
      jet_mul<C,karatsuba_multiply<> > (aux, t1, t2, n1, n2);
      jet_div (aux, p, n1+n2-1);
      jet_add (dest, aux, n1+n2-1);
      jet_derive (t1, n1);
      jet_coeff_derive (t2, n2);
      n1--; i++; p *= C(i);
    }
    DELETE_ARRAY (C, buf, (n1+n2)<<1);
  }
};

template<typename C>
class adjoint_helper<C,d_operator> {
public:
  static void indeterminate (C* dest) {
    dest[0]= C(0);
    dest[1]= C(-1);
  }
  static void mul (C* dest, const C* s1, const C* s2,
		   nat n1, nat n2, bool adj) {
    (void) adj;
    jet_mul<C,skew_multiply<d_operator> > (dest, s1, s2, n1, n2);
  }
};

/******************************************************************************
* The operator x d/dx
******************************************************************************/

class xd_operator {};

template<typename C> inline void
jet_coeff_xderive (C* dest, nat n) {
  while (n != 0) {
    *dest= xderive (*dest);
    dest++; n--;
  }
}

template<typename C>
class multiplier<C,skew_multiply<xd_operator>,C > {
public:
  static void mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    if (n1 == 0 || n2 == 0) return;
    INLINE_ARRAY (C, buf, (n1+n2)<<1);
    C* t1 = buf;
    C* t2 = t1 + n1;
    C* aux= t2 + n2;
    jet_clear (dest, n1+n2-1);
    jet_copy (t1, s1, n1);
    jet_copy (t2, s2, n2);
    int i=0;
    C p (1);
    while (n1 > 0) {
      jet_mul<C,karatsuba_multiply<> > (aux, t1, t2, n1, n2);
      jet_div (aux, p, n1+n2-1);
      jet_add (dest, aux, n1+n2-1);
      jet_derive (t1, n1);
      jet_coeff_xderive (t2, n2);
      n1--; i++; p *= C(i);
    }
    DELETE_ARRAY (C, buf, (n1+n2)<<1);
  }
};

template<typename C>
class adjoint_helper<C,xd_operator> {
public:
  static void indeterminate (C* dest) {
    dest[0]= C(-1);
    dest[1]= C(-1);
  }
  static void mul (C* dest, const C* s1, const C* s2,
		   nat n1, nat n2, bool adj) {
    (void) adj;
    jet_mul<C,skew_multiply<xd_operator> > (dest, s1, s2, n1, n2);
  }
};

/******************************************************************************
* The shift operators x |-> x + 1 and x |-> x - 1
******************************************************************************/

template<typename K> class S_operator {};
template<typename K> class T_operator {};

template<typename C, typename K> inline void
jet_coeff_shift (C* dest, const K& sh, nat n) {
  while (n != 0) {
    *dest= shift (*dest, sh);
    dest++; n--;
  }
}

template<typename C, typename K> void
mul_S_operator (C* dest, const C* s1, const C* s2,
		const K& sh, nat n1, nat n2)
{
  if (n1 == 0 || n2 == 0) return;
  INLINE_ARRAY (C, t2, n2);
  jet_clear (dest, n1+n2-1);
  jet_copy (t2, s2, n2);
  while (n1 != 0) {
    jet_mul_add (dest, *s1, t2, n2);
    jet_coeff_shift (t2, sh, n2);
    dest++; s1++; n1--;
  }
  DELETE_ARRAY (C, t2, n2);
}

template<typename C, typename K>
class multiplier<C,skew_multiply<S_operator<K> >,C > {
public:
  static void mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    mul_S_operator (dest, s1, s2, K(1), n1, n2);
  }
};

template<typename C, typename K>
class multiplier<C,skew_multiply<T_operator<K> >,C > {
public:
  static void mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    mul_S_operator (dest, s1, s2, K(-1), n1, n2);
  }
};

template<typename C, typename K>
class adjoint_helper<C,S_operator<K> > {
public:
  static void indeterminate (C* dest) {
    dest[0]= C(0);
    dest[1]= C(1);
  }
  static void mul (C* dest, const C* s1, const C* s2,
		   nat n1, nat n2, bool adj) {
    if (adj) jet_mul<C,skew_multiply<T_operator<K> > > (dest, s1, s2, n1, n2);
    else jet_mul<C,skew_multiply<S_operator<K> > > (dest, s1, s2, n1, n2);
  }
};

template<typename C, typename K>
class adjoint_helper<C,T_operator<K> > {
public:
  static void indeterminate (C* dest) {
    dest[0]= C(0);
    dest[1]= C(1);
  }
  static void mul (C* dest, const C* s1, const C* s2,
		   nat n1, nat n2, bool adj) {
    if (adj) jet_mul<C,skew_multiply<S_operator<K> > > (dest, s1, s2, n1, n2);
    else jet_mul<C,skew_multiply<T_operator<K> > > (dest, s1, s2, n1, n2);
  }
};

/******************************************************************************
* Alternative multiplication for shift operators with polynomial coefficients
* Can be faster for operators of high orders
******************************************************************************/

template<typename C> nat
jet_coeff_dim (const C* s, nat n) {
  nat m= 0;
  while (n != 0) {
    m= max (dim (*s), m);
    s++; n--;
  }
  return m;
}

template<typename C, typename K> void
mul_pol_S_operator (C* dest, const C* s1, const C* s2,
		    const K& sh, nat n1, nat n2)
{
  if (n1 == 0 || n2 == 0) return;
  INLINE_ARRAY (C, buf, (n1+n2)<<1);
  C* t1 = buf;
  C* t2 = t1 + n1;
  C* aux= t2 + n2;
  jet_clear (dest, n1+n2-1);
  jet_copy (t1, s1, n1);
  jet_copy (t2, s2, n2);
  int i= 0, r= jet_coeff_dim (t2, n2);
  C p (1), q (1);
  for (i=0; i<r; i++) {
    jet_mul<C,karatsuba_multiply<> > (aux, t1, t2, n1, n2);
    jet_mul (aux, p, n1+n2-1);
    jet_div (aux, q, n1+n2-1);
    jet_add (dest, aux, n1+n2-1);
    jet_xderive (t1, n1);
    jet_coeff_derive (t2, n2);
    p *= C(sh); q *= C(i+1);
  }
  DELETE_ARRAY (C, buf, (n1+n2)<<1);
}

// FIXME: we should find a criterium for taking the most efficient
// algorithm among mul_S_operator and mul_pol_S_operator

/*
template<typename K, typename V>
class multiplier<polynomial<K,V>,skew_multiply<S_operator<K> >,
                 polynomial<K,V> > {
public:
  typedef polynomial<K,V> C;
  static void mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    mul_pol_S_operator (dest, s1, s2, K(1), n1, n2);
  }
};

template<typename K, typename V>
class multiplier<polynomial<K,V>,skew_multiply<T_operator<K> >,
                 polynomial<K,V> > {
public:
  typedef polynomial<K,V> C;
  static void mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    mul_pol_S_operator (dest, s1, s2, K(-1), n1, n2);
  }
};
*/

/******************************************************************************
* Dichotomic computation of adjoints
******************************************************************************/

template<typename C, typename Op> void
jet_dicho_adjoint (C* dest, C* p, const C* s, nat n, bool adj) {
  // computes adjoint ((s, n)) and the n-th power of the skew indeterminate u
  if (n == 1) {
    *dest= *s;
    adjoint_helper<C,Op>::indeterminate (p);
  }
  else {
    nat h1= n >> 1, h2= n - h1;
    nat k1= h1 + 1, k2= h2 + 1;
    INLINE_ARRAY (C, c1, h1 + h2 + k1 + k2);
    C* c2= c1 + h1;
    C* p1= c2 + h2;
    C* p2= p1 + k1;
    jet_dicho_adjoint<C,Op> (c1, p1, s, h1, adj);
    jet_dicho_adjoint<C,Op> (c2, p2, s+h1, h2, adj);
    adjoint_helper<C,Op>::mul (dest, p1, c2, k1, h2, !adj);
    jet_add (dest, c1, h1);
    adjoint_helper<C,Op>::mul (p, p1, p2, k1, k2, !adj);
    DELETE_ARRAY (C, c1, h1 + h2 + k1 + k2);
  }
}

template<typename C, typename Op> void
jet_dicho_adjoint (C* dest, const C* s, nat n, bool adj) {
  // correct for operator rings C[u] with deg (u^*) = 1,
  // which implies deg (L^*) = deg (L) for all L
  assert (n != 0);
  if (n == 1) *dest= *s;
  else {
    nat h1= n >> 1, h2= n - h1;
    nat k1= h1 + 1;
    INLINE_ARRAY (C, c1, h1 + h2 + k1);
    C* c2= c1 + h1;
    C* p1= c2 + h2;
    jet_dicho_adjoint<C,Op> (c1, p1, s, h1, adj);
    jet_dicho_adjoint<C,Op> (c2, s+h1, h2, adj);
    adjoint_helper<C,Op>::mul (dest, p1, c2, k1, h2, !adj);
    jet_add (dest, c1, h1);
    DELETE_ARRAY (C, c1, h1 + h2 + k1);
  }
}

template<typename C, typename Op> void
jet_dicho_adjoint (C* dest, nat n, bool adj) {
  INLINE_ARRAY (C, aux, n);
  jet_copy (aux, dest, n);
  jet_dicho_adjoint<C,Op> (dest, aux, n, adj);
  DELETE_ARRAY (C, aux, n);
}

/******************************************************************************
* Skew dichotomic division
******************************************************************************/

template<typename C, typename Op> void
jet_skew_quo_rem (C* dest, C* s1, const C* s2,
		  nat n1, nat n2, bool left, bool adj)
{
  // left  division of A by B (i.e. B^(-1) A) yields A = B Q + R
  // right division of A by B (i.e. A B^(-1)) yields A = Q B + R
  // adj indicates whether we are using the adjoint operator of Op
  assert (n2 != 0);
  if (n1 < n2);
  else if (n1 == 1) {
    *dest= skew_quo (*s1, *s2, left);
    if (left) *s1 -= (*s2) * (*dest);
    else *s1 -= (*dest) * (*s2);
  }
  else {
    nat l= n1-n2+1;
    if (n2 > l) {
      if (left) {
	nat d= n2-l;
	INLINE_ARRAY (C, S2, n2 + l+d-1);
	C* buf= S2+ n2;
	jet_dicho_adjoint<C,Op> (s1, n1, adj);
	jet_dicho_adjoint<C,Op> (S2, s2, n2, adj);
	jet_skew_quo_rem<C,Op> (dest, s1+d, S2+d, n1-d, l, false, !adj);
	adjoint_helper<C,Op>::mul (buf, dest, S2, l, d, !adj);
	jet_sub (s1, buf, l+d-1);
	jet_dicho_adjoint<C,Op> (dest, l, !adj);
	jet_dicho_adjoint<C,Op> (s1, n1, !adj);
	DELETE_ARRAY (C, S2, n2 + l+d-1);
      }
      else {
	nat d= n2-l;
	INLINE_ARRAY (C, buf, l+d-1);
	jet_skew_quo_rem<C,Op> (dest, s1+d, s2+d, n1-d, l, false, adj);
	adjoint_helper<C,Op>::mul (buf, dest, s2, l, d, adj);
	jet_sub (s1, buf, l+d-1);
	DELETE_ARRAY (C, buf, l+d-1);
      }
    }
    else {
      if (left) {
	nat h= (n2+1) >> 1;
	nat d= l-h;
	jet_skew_quo_rem<C,Op> (dest+d, s1+d, s2, n1-d, n2, true, adj);
	jet_skew_quo_rem<C,Op> (dest, s1, s2, n1-h, n2, true, adj);
      }
      else {
	nat h= (n2+1) >> 1;
	nat d= l-h;
	INLINE_ARRAY (C, S2, n2);
	jet_dicho_adjoint<C,Op> (s1, n1, adj);
	jet_dicho_adjoint<C,Op> (S2, s2, n2, adj);
	jet_skew_quo_rem<C,Op> (dest+d, s1+d, S2, n1-d, n2, true, !adj);
	jet_skew_quo_rem<C,Op> (dest, s1, S2, n1-h, n2, true, !adj);
	jet_dicho_adjoint<C,Op> (dest, l, !adj);
	jet_dicho_adjoint<C,Op> (s1, n1, !adj);
	DELETE_ARRAY (C, S2, n2);
      }
    }
  }
}

template<typename C, typename Op>
class divisor<C,skew_multiply<Op> > {
public:
  static void quo_rem (C* dest, C* s1, const C* s2, nat n1, nat n2) {
    return jet_skew_quo_rem<C,Op> (dest, s1, s2, n1, n2, false, false); }
};

} // namespace mmx
#endif //__MMX__SKEW_JET__HPP
