
/******************************************************************************
* MODULE     : ball.hpp
* DESCRIPTION: Balls
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_BALL_HPP
#define __MMX_BALL_HPP
#include "rational.hpp"
#include "type_props.hpp"
namespace mmx {
#define TMPL_DEF template<typename C,typename R,typename V=lipschitz_ball>
#define TMPL template<typename C,typename R,typename V>
#define Ball ball<C,R,V>
TMPL class ball;
TMPL inline C center (const Ball& z);
TMPL inline R radius (const Ball& z);
TMPL inline Ball operator - (const Ball& z1, const Ball& z2);

template<typename C> inline C
add_U (const C& c1, const C& c2, const C& c3) {
  return add_U (add_U (c1, c2), c3);
}

template<typename C> inline C
mul_U (const C& c1, const C& c2, const C& c3) {
  return mul_U (mul_U (c1, c2), c3);
}

/******************************************************************************
* Ball variants
******************************************************************************/

class lipschitz_ball {
public:
  template<typename C,typename R> static inline void
  normalize (C& c, R& r) { r= add_U (r, rnd_error (c)); }
};

class truncated_lipschitz_ball {
public:
  template<typename C,typename R> static void
  normalize (C& c, R& r) {
    if (is_number (c) && is_number (r)) {
      if (step (c) < step (r)) {
	int p= precision (c) - mpfr_prec_round_up (step (r) - step (c));
	c= copy (c, (nat) max ((int) DEFAULT_LOW, p));
      }
    }
    else c= copy (c, DEFAULT_LOW);
    r= add_U (r, rnd_error (c));
  }
};

/******************************************************************************
* Rounding routines for balls
******************************************************************************/

template<typename T,typename F>
class ball_rounding_helper {
public:
  static T cv_D (const F& x);
  static T cv_U (const F& x);
  static T cv_abs_D (const F& x);
  static T cv_abs_U (const F& x);
};

template<typename T,typename F> inline T
convert_D (const F& x) {
  return ball_rounding_helper<T,F>::cv_D (x); }

template<typename T,typename F> inline T
convert_U (const F& x) {
  return ball_rounding_helper<T,F>::cv_U (x); }

template<typename T,typename F> inline T
convert_abs_D (const F& x) {
  return ball_rounding_helper<T,F>::cv_abs_D (x); }

template<typename T,typename F> inline T
convert_abs_U (const F& x) {
  return ball_rounding_helper<T,F>::cv_abs_U (x); }

/******************************************************************************
* The ball type
******************************************************************************/

TMPL_DEF
class ball {
MMX_ALLOCATORS;
private:
  C c;
  R r;

public:
  inline void normalize () { V::normalize (c, r); }
  inline ball (): c (0), r (0) {}
  template<typename T> ball (const T& c2):
    c (c2), r (0) { normalize (); }
  template<typename CT,typename RT> inline ball (const ball<CT,RT,V>& z):
    c (center (z)), r (radius (z)) { normalize (); }
  inline ball (const C& x): c (copy (x)), r (0) { normalize (); }
  inline ball (const C& c2, const R& r2):
    c (c2), r (r2) { normalize (); }
  inline ball (const C& c2, const R& r2, bool flag):
    c (c2), r (r2) { if (flag) normalize (); }
  inline ball (const Real_type(Ball)& x, const Real_type(Ball)& y):
    c (gaussian (center (x), center (y))),
    r (hypot_U (radius (x), radius (y))) { normalize (); }
  friend C center LESSGTR (const Ball& z);
  friend R radius LESSGTR (const Ball& z);

  inline Ball& operator <<= (int shift);
  inline Ball& operator >>= (int shift);
};

STYPE_TO_TYPE(TMPL,center_type,Ball,C);
STYPE_TO_TYPE(TMPL,radius_type,Ball,R);

TMPL
struct constant_helper<Ball > {
  static inline Ball nan () {
    return Ball (Nan (C), 0, false); }
  static inline Ball plus_infinity () {
    return Ball (Plus_infinity (C), 0, false); }
  static inline Ball minus_infinity () {
    return Ball (Minus_infinity (C), 0, false); }
  static inline Ball infinity () {
    return Ball (Infinity (C), 0, false); }
  static inline Ball fuzz () {
    return Ball (0, Plus_infinity (R), false); }
  static inline Ball log2 () { return Ball (Log2 (C), 0); }
  static inline Ball pi () { return Ball (Pi (C), 0); }
  static inline Ball euler () { return Ball (Euler (C), 0); }
  static inline Ball imaginary () {
    return Ball (Real_type(Ball) (0), Real_type(Ball) (1)); }
};

TMPL
struct make_ball_helper<Ball,C,R> {
  static inline Ball val (const C& c, const R& r) {
    return Ball (c, r); }
};

/******************************************************************************
* Basic routines
******************************************************************************/

TMPL inline C center (const Ball& z) {
  return z.c; }
TMPL inline R radius (const Ball& z) {
  return z.r; }
TMPL inline C lower (const Ball& z) {
  return var_sub_D (center (z), radius (z)); }
TMPL inline C upper (const Ball& z) {
  return var_add_U (center (z), radius (z)); }

TMPL inline R abs_D (const Ball& z) {
  return sub_D (convert_abs_D<R> (center(z)), radius(z)); }
TMPL inline R abs_U (const Ball& z) {
  return add_U (convert_abs_U<R> (center(z)), radius(z)); }
TMPL inline R bnd_D (const Ball& z) {
  return sub_D (convert_D<R> (center(z)), radius(z)); }
TMPL inline R bnd_U (const Ball& z) {
  return add_U (convert_U<R> (center(z)), radius(z)); }

TMPL inline bool is_exact_zero (const Ball& z) {
  return is_exact_zero (center (z)) && is_exact_zero (radius (z)); }
TMPL inline bool is_number (const Ball& z) {
  return is_number (center (z)) && is_number (radius (z)); }
TMPL inline bool is_infinity (const Ball& z) {
  return is_infinity (center (z)) && is_number (radius (z)); }
TMPL inline bool is_fuzz (const Ball& z) {
  return !is_nan (center (z)) && is_infinity (radius (z)); }
TMPL inline bool is_nan (const Ball& z) {
  return is_nan (center (z)) || is_nan (radius (z)); }

TMPL inline nat
hash (const Ball& z) {
  nat h= hash (center (z));
  return (h<<1) ^ (h<<3) ^ (h<<7) ^ (h>>29) ^ hash (radius (z));
}

TMPL generic
flatten (const Ball& z) {
  if (is_nan (z)) return Nan (generic);
  if (is_fuzz (z)) return Fuzz (generic);
  if (is_exact_zero (z)) return 0;
  return flatten (lower (z), upper (z));
  // return flatten ("ball", center (z), radius (z));
}

TMPL generic
flatten (const Ball& z, nat dd) {
  if (is_nan (z)) return Nan (generic);
  if (is_fuzz (z)) return Fuzz (generic);
  if (is_exact_zero (z)) return 0;
  return flatten (lower (z), upper (z), dd);
}

PRINT_SUGAR(TMPL,Ball)

/******************************************************************************
* Comparisons
******************************************************************************/

TMPL bool
hard_eq (const Ball& z1, const Ball& z2) {
  return hard_eq (center (z1), center (z2)) &&
         hard_eq (radius (z1), radius (z2));
}

TMPL bool
hard_neq (const Ball& z1, const Ball& z2) {
  return hard_neq (center (z1), center (z2)) ||
         hard_neq (radius (z1), radius (z2));
}

TMPL bool
eq (const Ball& z1, const Ball& z2) {
  return eq (center (z1), center (z2)) && eq (radius (z1), radius (z2));
}

TMPL bool
neq (const Ball& z1, const Ball& z2) {
  return neq (center (z1), center (z2)) || neq (radius (z1), radius (z2));
}

TMPL inline bool
is_zero (const Ball& z) {
  return !is_nan (center (z)) && !is_nan (radius (z)) &&
         abs (center (z)) <= radius (z);
}

TMPL inline bool
is_non_zero (const Ball& z) {
  return is_nan (center (z)) || is_nan (radius (z)) ||
         abs (center (z)) > radius (z);
}

TMPL inline bool
is_negative_or_zero (const Ball& z) {
  return !is_nan (center (z)) && center (z) <= radius (z);
}

TMPL inline bool
is_negative (const Ball& z) {
  return !is_nan (center (z)) && center (z) < -radius (z);
}

TMPL inline bool
is_positive_or_zero (const Ball& z) {
  return !is_nan (center (z)) && center (z) >= -radius (z);
}

TMPL inline bool
is_positive (const Ball& z) {
  return !is_nan (center (z)) && center (z) > radius (z);
}

TMPL bool
operator == (const Ball& z1, const Ball& z2) {
  if (is_nan (z1) || is_nan (z2)) return is_nan (z1) && is_nan (z2);
  Ball d= z1 - z2;
  return abs (center (d)) <= radius (d);
}

TMPL bool
operator != (const Ball& z1, const Ball& z2) {
  if (is_nan (z1) || is_nan (z2)) return !(is_nan (z1) && is_nan (z2));
  Ball d= z1 - z2;
  return abs (center (d)) > radius (d);
}

TMPL inline bool
operator <= (const Ball& z1, const Ball& z2) {
  return is_negative_or_zero (z1 - z2);
}

TMPL inline bool
operator < (const Ball& z1, const Ball& z2) {
  return is_negative (z1 - z2);
}

TMPL inline bool
operator >= (const Ball& z1, const Ball& z2) {
  return is_positive_or_zero (z1 - z2);
}

TMPL inline bool
operator > (const Ball& z1, const Ball& z2) {
  return is_positive (z1 - z2);
}

COMPARE_INT_SUGAR(TMPL,Ball,Ball);

/******************************************************************************
* Other order related functions
******************************************************************************/

TMPL inline Ball
abs (const Ball& z) {
  return Ball (abs (center (z)), radius (z), false);
}

TMPL Ball
min (const Ball& z1, const Ball& z2) {
  if (z1 < z2) return z1;
  else if (z2 < z1) return z2;
  else return Ball (min (center (z1), center (z2)),
		    max (radius (z1), radius (z2)),
		    false);
}

TMPL Ball
max (const Ball& z1, const Ball& z2) {
  if (z1 > z2) return z1;
  else if (z2 > z1) return z2;
  else return Ball (max (center (z1), center (z2)),
		    max (radius (z1), radius (z2)),
		    false);
}

/******************************************************************************
* Basic arithmetic
******************************************************************************/

TMPL Ball
operator - (const Ball& z) {
  if (is_number (z) || is_infinity (z))
    return Ball (-center (z), radius (z), false);
  return z;
}

TMPL Ball
operator + (const Ball& z1, const Ball& z2) {
  if (is_number (z1) && is_number (z2))
    return Ball (center (z1) + center (z2),
		 add_U (radius (z1), radius (z2)));
  if (is_nan (z1) || is_nan (z2)) return Nan (Ball);
  if (is_fuzz (z1) || is_fuzz (z2)) return Fuzz (Ball);
  if (is_infinity (z1) && is_infinity (z2) &&
      var_sign (center (z1)) == var_sign (-center (z2)))
    return Nan (Ball);
  return is_infinity (z1)? z1: z2;
}

TMPL Ball
operator - (const Ball& z1, const Ball& z2) {
  if (is_number (z1) && is_number (z2))
    return Ball (center (z1) - center (z2),
		 add_U (radius (z1), radius (z2)));
  if (is_nan (z1) || is_nan (z2)) return Nan (Ball);
  if (is_fuzz (z1) || is_fuzz (z2)) return Fuzz (Ball);
  if (is_infinity (z1) && is_infinity (z2) &&
      var_sign (center (z1)) == var_sign (center (z2)))
    return Nan (Ball);
  return is_infinity (z1)? z1: -z2;
}

TMPL Ball
operator * (const Ball& z1, const Ball& z2) {
  if (is_number (z1) && is_number (z2)) {
    C c= center (z1) * center (z2);
    if (is_nan (c)) return Fuzz(Ball);
    return Ball (c, add_U (mul_U (abs_U (z1), radius (z2)),
			   mul_U (radius (z1), abs_U (z2))));
  }
  if (is_nan (z1) || is_nan (z2)) return Nan (Ball);
  if (is_fuzz (z1) || is_fuzz (z2)) return Fuzz (Ball);
  if (is_infinity (z1)) {
    if (is_zero (z2)) return Fuzz (Ball);
    if (var_sign (z1) == var_sign (-z1)) return Infinity (Ball);
    return Ball (center (z1) * center (z2), 0, false);
  }
  else {
    if (is_zero (z1)) return Fuzz (Ball);
    if (var_sign (z2) == var_sign (-z2)) return Infinity (Ball);
    return Ball (center (z1) * center (z2), 0, false);
  }
}

TMPL inline Ball
square (const Ball& z) {
  if (is_number (z)) {
    C c= square (center (z));
    if (is_nan (c)) return Fuzz(Ball);
    return Ball (square (center (z)),
		 mul_U (abs_U (z), radius (z)) << 1);
  }
  if (is_infinity (z)) {
    if (var_sign (z) == var_sign (-z)) return Infinity (Ball);
    return Ball (square (center (z)), 0, false);
  }
  return z;
}

TMPL Ball
operator / (const Ball& z1, const Ball& z2) {
  if (is_zero (z2)) return Nan (Ball);
  if (is_number (z1) && is_number (z2))
    return z1 * Ball (1 / center (z2),
		      div_U (radius (z2), square_U (abs_D (z2))));
  if (is_nan (z1) || is_nan (z2)) return Nan (Ball);
  if (is_fuzz (z1) || is_fuzz (z2)) return Fuzz (Ball);
  if (is_infinity (z1)) {
    if (is_infinity (z2)) return Fuzz (Ball);
    if (var_sign (z1) == var_sign (-z1)) return Infinity (Ball);
    return Ball (center (z1) / center (z2), 0, false);
  }
  else return 0;
}

TMPL inline Ball
operator << (const Ball& z, int shift) {
  return Ball (center (z) << shift, radius (z) << shift, false);
}

TMPL inline Ball
operator >> (const Ball& z, int shift) {
  return Ball (center (z) >> shift, radius (z) >> shift, false);
}

ARITH_INT_SUGAR(TMPL,Ball,Ball);

/******************************************************************************
* In place arithmetic operations
******************************************************************************/

TMPL inline Ball&
Ball::operator <<= (int shift) {
  c <<= shift;
  r <<= shift;
  return *this;
}

TMPL inline Ball&
Ball::operator >>= (int shift) {
  c >>= shift;
  r >>= shift;
  return *this;
}

/******************************************************************************
* Elementary functions
******************************************************************************/

TMPL Ball
sqrt (const Ball& z) {
  if (is_negative_or_zero (z)) return Nan (Ball);
  return Ball (sqrt (center (z)),
	       div_U (radius (z) >> 1, sqrt_D (bnd_D (z))));
}

TMPL Ball
exp (const Ball& z) {
  return Ball (exp (center (z)),
	       mul_U (exp_U (bnd_U (z)), radius (z)));
}

TMPL Ball
log (const Ball& z) {
  if (is_negative_or_zero (z)) return Nan (Ball);
  return Ball (log (center (z)),
	       div_U (radius (z), bnd_D (z)));
}

TMPL inline Ball
pow (const Ball& z, const Ball& y) {
  return exp (y * log (z));
}

TMPL Ball
cos (const Ball& z) {
  if (radius (z) >= R (3.14159))
    return Ball (cos (center (z)), radius (z));
  else {
    R u1= sin (bnd_D (z));
    R u2= sin (bnd_U (z));
    if (sign (u1) * sign (u2) != 1)
      return Ball (cos (center (z)), radius (z));
    return Ball (cos (center (z)),
		 mul_U (max (abs (u1), abs (u2)), radius (z)));
  }
}

TMPL Ball
sin (const Ball& z) {
  if (radius (z) >= R (3.14159))
    return Ball (sin (center (z)), radius (z));
  else {
    R u1= cos (bnd_D (z));
    R u2= cos (bnd_U (z));
    if (sign (u1) * sign (u2) != 1)
      return Ball (sin (center (z)), radius (z));
    return Ball (sin (center (z)),
		 mul_U (max (abs (u1), abs (u2)), radius (z)));
  }
}

TMPL inline Ball
cosh (const Ball& z) {
  return Ball (cosh (center (z)),
	       mul_U (sinh_U (bnd_U (abs (z))), radius (z)));
}

TMPL inline Ball
sinh (const Ball& z) {
  return Ball (sinh (center (z)),
	       mul_U (cosh_U (bnd_U (abs (z))), radius (z)));
}

TMPL Ball
acos (const Ball& z) {
  Ball u= 1 - square (z);
  R v= bnd_D (u);
  if (v <= 0) return Nan (Ball);
  return Ball (acos (center (z)), div_U (radius (z), sqrt_D (v)));
}

TMPL Ball
asin (const Ball& z) {
  Ball u= 1 - square (z);
  R v= bnd_D (u);
  if (v <= 0) return Nan (Ball);
  return Ball (asin (center (z)), div_U (radius (z), sqrt_D (v)));
}

TMPL Ball
atan (const Ball& z) {
  R u= bnd_D (z);
  R v= add_D (R (1), square_D (u));
  if (u <= 0) return Ball (atan (center (z)), radius (z));
  else return Ball (atan (center (z)), div_U (radius (z), v));
}

TMPL Ball
atan2 (const Ball& y, const Ball& x) {
  Ball z= square (x) + square (y);
  if (z <= 0) return Nan (Ball);
  R b = div_U (R (1), bnd_D (z));
  R dx= mul_U (bnd_U (abs (y)), b);
  R dy= mul_U (bnd_U (abs (x)), b);
  return Ball (atan2 (center (x), center (y)),
	       add_U (mul_U (dx, radius (x)),
		      mul_U (dy, radius (y))));
}

#undef TMPL_DEF
#undef TMPL
#undef Ball
} // namespace mmx
#endif // __MMX_BALL_HPP
