
/******************************************************************************
* MODULE     : rounded_floating.hpp
* DESCRIPTION: Floating point numbers with explicit rounding
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_ROUNDED_FLOATING_HPP
#define __MMX_ROUNDED_FLOATING_HPP
#include "floating.hpp"
namespace mmx {
#define TMPL template<typename V>
#define Floating floating<V>
#define Floating_rep floating_rep<V>

/******************************************************************************
* Constructors
******************************************************************************/

TMPL
struct rounded_constructor_helper<Floating > {
  template<class C> static inline Floating rnd_D (const C& x) {
    return as_floating<V> (floating<rnd_floating<V,GMP_RNDD> > (x)); }
  template<class C> static inline Floating rnd_U (const C& x) {
    return as_floating<V> (floating<rnd_floating<V,GMP_RNDU> > (x)); }
};

/******************************************************************************
* Basic arithmetic
******************************************************************************/

TMPL inline Floating add_D (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_add (*r, *x1, *x2, GMP_RNDD); return r; }
TMPL inline Floating sub_D (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_sub (*r, *x1, *x2, GMP_RNDD); return r; }
TMPL inline Floating mul_D (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_mul (*r, *x1, *x2, GMP_RNDD); return r; }
TMPL inline Floating div_D (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_div (*r, *x1, *x2, GMP_RNDD); return r; }
TMPL inline Floating square_D (const Floating& x) {
  Floating r (*x); mpfr_mul (*r, *x, *x, GMP_RNDD); return r; }

TMPL inline Floating add_U (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_add (*r, *x1, *x2, GMP_RNDU); return r; }
TMPL inline Floating sub_U (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_sub (*r, *x1, *x2, GMP_RNDU); return r; }
TMPL inline Floating mul_U (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_mul (*r, *x1, *x2, GMP_RNDU); return r; }
TMPL inline Floating div_U (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_div (*r, *x1, *x2, GMP_RNDU); return r; }
TMPL inline Floating square_U (const Floating& x) {
  Floating r (*x); mpfr_mul (*r, *x, *x, GMP_RNDU); return r; }

template<typename V,typename W> inline floating<V>
add_U (const floating<V>& x1, const floating<W>& x2) {
  floating<V> r (*x1, *x2); mpfr_add (*r, *x1, *x2, GMP_RNDD); return r; }
template<typename V,typename W> inline floating<V>
var_sub_D (const floating<V>& x1, const floating<W>& x2) {
  floating<V> r (new floating_rep<V> ((nat) mpfr_get_prec (*x1)));
  mpfr_sub (*r, *x1, *x2, GMP_RNDD);
  return r; }
template<typename V,typename W> inline floating<V>
var_add_U (const floating<V>& x1, const floating<W>& x2) {
  floating<V> r (new floating_rep<V> ((nat) mpfr_get_prec (*x1)));
  mpfr_add (*r, *x1, *x2, GMP_RNDU);
  return r; }

/******************************************************************************
* Elementary functions
******************************************************************************/

TMPL inline Floating sqrt_D (const Floating& x1) {
  Floating r (*x1); mpfr_sqrt (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating cbrt_D (const Floating& x1) {
  Floating r (*x1); mpfr_cbrt (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating hypot_D (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_hypot (*r, *x1, *x2, GMP_RNDD); return r; }
TMPL inline Floating exp_D (const Floating& x1) {
  Floating r (*x1); mpfr_exp (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating exp2_D (const Floating& x1) {
  Floating r (*x1); mpfr_exp2 (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating log_D (const Floating& x1) {
  Floating r (*x1); mpfr_log (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating log2_D (const Floating& x1) {
  Floating r (*x1); mpfr_log2 (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating log10_D (const Floating& x1) {
  Floating r (*x1); mpfr_log10 (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating pow_D (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_pow (*r, *x1, *x2, GMP_RNDD); return r; }
TMPL inline Floating cos_D (const Floating& x1) {
  Floating r (*x1); mpfr_cos (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating sin_D (const Floating& x1) {
  Floating r (*x1); mpfr_sin (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating tan_D (const Floating& x1) {
  Floating r (*x1); mpfr_tan (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating acos_D (const Floating& x1) {
  Floating r (*x1); mpfr_acos (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating asin_D (const Floating& x1) {
  Floating r (*x1); mpfr_asin (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating atan_D (const Floating& x1) {
  Floating r (*x1); mpfr_atan (*r, *x1, GMP_RNDD); return r; }
//TMPL inline Floating atan2_D (const Floating& x1, const Floating& x2) {
//Floating r (*x1, *x2); mpfr_atan2 (*r, *x1, *x2, GMP_RNDD); return r; }
TMPL inline Floating cosh_D (const Floating& x1) {
  Floating r (*x1); mpfr_cosh (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating sinh_D (const Floating& x1) {
  Floating r (*x1); mpfr_sinh (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating tanh_D (const Floating& x1) {
  Floating r (*x1); mpfr_tanh (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating acosh_D (const Floating& x1) {
  Floating r (*x1); mpfr_acosh (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating asinh_D (const Floating& x1) {
  Floating r (*x1); mpfr_asinh (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating atanh_D (const Floating& x1) {
  Floating r (*x1); mpfr_atanh (*r, *x1, GMP_RNDD); return r; }

TMPL inline Floating sqrt_U (const Floating& x1) {
  Floating r (*x1); mpfr_sqrt (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating cbrt_U (const Floating& x1) {
  Floating r (*x1); mpfr_cbrt (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating hypot_U (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_hypot (*r, *x1, *x2, GMP_RNDU); return r; }
TMPL inline Floating exp_U (const Floating& x1) {
  Floating r (*x1); mpfr_exp (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating exp2_U (const Floating& x1) {
  Floating r (*x1); mpfr_exp2 (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating log_U (const Floating& x1) {
  Floating r (*x1); mpfr_log (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating log2_U (const Floating& x1) {
  Floating r (*x1); mpfr_log2 (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating log10_U (const Floating& x1) {
  Floating r (*x1); mpfr_log10 (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating pow_U (const Floating& x1, const Floating& x2) {
  Floating r (*x1, *x2); mpfr_pow (*r, *x1, *x2, GMP_RNDU); return r; }
TMPL inline Floating cos_U (const Floating& x1) {
  Floating r (*x1); mpfr_cos (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating sin_U (const Floating& x1) {
  Floating r (*x1); mpfr_sin (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating tan_U (const Floating& x1) {
  Floating r (*x1); mpfr_tan (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating acos_U (const Floating& x1) {
  Floating r (*x1); mpfr_acos (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating asin_U (const Floating& x1) {
  Floating r (*x1); mpfr_asin (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating atan_U (const Floating& x1) {
  Floating r (*x1); mpfr_atan (*r, *x1, GMP_RNDU); return r; }
//TMPL inline Floating atan2_U (const Floating& x1, const Floating& x2) {
//Floating r (*x1, *x2); mpfr_atan2 (*r, *x1, *x2, GMP_RNDU); return r; }
TMPL inline Floating cosh_U (const Floating& x1) {
  Floating r (*x1); mpfr_cosh (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating sinh_U (const Floating& x1) {
  Floating r (*x1); mpfr_sinh (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating tanh_U (const Floating& x1) {
  Floating r (*x1); mpfr_tanh (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating acosh_U (const Floating& x1) {
  Floating r (*x1); mpfr_acosh (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating asinh_U (const Floating& x1) {
  Floating r (*x1); mpfr_asinh (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating atanh_U (const Floating& x1) {
  Floating r (*x1); mpfr_atanh (*r, *x1, GMP_RNDU); return r; }

/******************************************************************************
* Special functions
******************************************************************************/

TMPL inline Floating gamma_D (const Floating& x1) {
  Floating r (*x1); mpfr_gamma (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating zeta_D (const Floating& x1) {
  Floating r (*x1); mpfr_zeta (*r, *x1, GMP_RNDD); return r; }
TMPL inline Floating erf_D (const Floating& x1) {
  Floating r (*x1); mpfr_erf (*r, *x1, GMP_RNDD); return r; }

TMPL inline Floating gamma_U (const Floating& x1) {
  Floating r (*x1); mpfr_gamma (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating zeta_U (const Floating& x1) {
  Floating r (*x1); mpfr_zeta (*r, *x1, GMP_RNDU); return r; }
TMPL inline Floating erf_U (const Floating& x1) {
  Floating r (*x1); mpfr_erf (*r, *x1, GMP_RNDU); return r; }

#undef TMPL
#undef Floating
#undef Floating_rep
}
#endif // __MMX_FLOATING_HPP
