
/******************************************************************************
* MODULE     : simple_glue.cpp
* DESCRIPTION: Basic simplifiation for symbolic expressions
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "base_evaluator.hpp"
#include "glue.hpp"
#include "rational.hpp"
namespace mmx {

generic simple_neg (const generic& g);
generic simple_add (const generic& g1, const generic& g2);
generic simple_sub (const generic& g1, const generic& g2);
generic simple_mul (const generic& g1, const generic& g2);
generic simple_div (const generic& g1, const generic& g2);
generic simple_pow (const generic& g1, const generic& g2);

/******************************************************************************
* Additive arithmetic with simplifications
******************************************************************************/

static void
signed_decompose (const generic& g, generic& abs_g, int& sgn_g) {
  if (is<literal> (g)) {
    string s= literal_to_string (g);
    if (N(s) > 1 && s[0] == '-' &&
	(s[1] == '.' || (s[1] >= '0' && s[1] <= '9'))) {
      abs_g= generic (s (1, N(s)));
      sgn_g= -1;
    }
    else {
      abs_g= g;
      sgn_g= 1;
    }
  }
  else if (is_func (g, MMX_MINUS, 1)) {
    signed_decompose (g[1], abs_g, sgn_g);
    sgn_g= -sgn_g;
  }
  else if (is_func (g, MMX_TIMES, 2) || is_func (g, MMX_OVER, 2)) {
    generic a1, a2;
    int     s1, s2;
    signed_decompose (g[1], a1, s1);
    signed_decompose (g[2], a2, s2);
    if (is_func (g, MMX_TIMES, 2)) abs_g= a1 * a2;
    else abs_g= a1 / a2;
    sgn_g= s1 * s2;
  }
  else {
    abs_g= g;
    sgn_g= 1;
  }
}

generic
simple_neg (const generic& g) {
  if (eq (g, 0)) return g;
  if (is_func (g, MMX_MINUS, 1))
    return g[1];
  if (is_func (g, MMX_PLUS, 2))
    return simple_sub (simple_neg (g[1]), g[2]);
  if (is_func (g, MMX_MINUS, 2))
    return simple_add (simple_neg (g[1]), g[2]);
  generic abs_g; int sgn_g;
  signed_decompose (g, abs_g, sgn_g);
  if (sgn_g > 0) return apply (MMX_MINUS, abs_g);
  else return abs_g;
}

generic
simple_add (const generic& g1, const generic& g2) {
  if (eq (g1, 0)) return g2;
  if (eq (g2, 0)) return g1;
  if (is_func (g2, MMX_MINUS, 1))
    return simple_sub (g1, g2[1]);
  if (is_func (g2, MMX_PLUS, 2))
    return simple_add (simple_add (g1, g2[1]), g2[2]);
  if (is_func (g2, MMX_MINUS, 2))
    return simple_sub (simple_add (g1, g2[1]), g2[2]);
  generic abs_g2; int sgn_g2;
  signed_decompose (g2, abs_g2, sgn_g2);
  if (sgn_g2 < 0) return apply (MMX_MINUS, g1, abs_g2);
  else return apply (MMX_PLUS, g1, abs_g2);
}

generic
simple_sub (const generic& g1, const generic& g2) {
  if (eq (g2, 0)) return g1;
  if (is_func (g2, MMX_MINUS, 1))
    return simple_add (g1, g2[1]);
  if (is_func (g2, MMX_PLUS, 2))
    return simple_sub (simple_sub (g1, g2[1]), g2[2]);
  if (is_func (g2, MMX_MINUS, 2))
    return simple_add (simple_sub (g1, g2[1]), g2[2]);
  generic abs_g2; int sgn_g2;
  signed_decompose (g2, abs_g2, sgn_g2);
  if (eq (g1, 0)) {
    if (sgn_g2 > 0) return apply (MMX_MINUS, abs_g2);
    else return abs_g2;    
  }
  else {
    if (sgn_g2 > 0) return apply (MMX_MINUS, g1, abs_g2);
    else return apply (MMX_PLUS, g1, abs_g2);
  }
}

static generic
migrate_sub (const generic& g, const generic& h) {
  if (is_func (g, MMX_MINUS, 2))
    return apply (MMX_MINUS, migrate_sub (g[1], h), g[2]);
  generic abs_g; int sgn_g;
  signed_decompose (g, abs_g, sgn_g);
  return apply (MMX_MINUS, h, abs_g);
}

static generic
migrate_sub (const generic& g, nat found) {
  if (found == 0) return migrate_sub (g[1], g[2]);
  return (g[0]) (migrate_sub (g[1], found-1), g[2]);
}

generic
migrate_negate (const generic& g, nat maximal) {
  int i= 0, found= -1;
  generic h= g;
  while (is_func (h, MMX_PLUS, 2) || is_func (h, MMX_MINUS, 2)) {
    if (is_func (h, MMX_PLUS, 2)) found= i;
    if (maximal == 0) return g;
    h= h[1];
    maximal--;
    i++;
  }
  if (found == -1) return g;
  generic abs_h; int sgn_h;
  signed_decompose (h, abs_h, sgn_h);
  if (sgn_h >= 0) return g;
  return migrate_sub (g, found);
}

/******************************************************************************
* Multiplicative arithmetic with simplifications
******************************************************************************/

static bool frac_flag= false;

bool
set_frac_flag (bool new_val) {
  bool ret= frac_flag;
  frac_flag= new_val;
  return ret;
}

generic
simple_mul (const generic& g1, const generic& g2) {
  if (frac_flag && is_func (g1, MMX_OVER, 2)) {
    if (is_func (g2, MMX_OVER, 2))
      return apply (MMX_OVER, simple_mul (g1[1], g2[1]),
		              simple_mul (g1[2], g2[2]));
    else return apply (MMX_OVER, simple_mul (g1[1], g2), g1[2]);
  }
  if (frac_flag && is_func (g2, MMX_OVER, 2))
    return apply (MMX_OVER, simple_mul (g1, g2[1]), g2[2]);
  if (eq (g1, 0)) return 0;
  if (eq (g2, 0)) return 0;
  if (eq (g1, 1)) return g2;
  if (eq (g2, 1)) return g1;
  if (eq (g1, -1)) return simple_neg (g2);
  if (eq (g2, -1)) return simple_neg (g1);
  return apply (MMX_TIMES, g1, g2);
}

generic
simple_div (const generic& g1, const generic& g2) {
  if (frac_flag && is_func (g2, MMX_OVER, 2))
    return simple_mul (g1, apply (MMX_OVER, g2[2], g2[1]));
  if (eq (g2, 1)) return g1;
  if (eq (g2, -1)) return simple_neg (g1);
  if (frac_flag) return simple_mul (g1, apply (MMX_OVER, 1, g2));
  else return apply (MMX_OVER, g1, g2);
}

generic
simple_pow (const generic& g1, const generic& g2) {
  if (eq (g1, 0)) return 0;
  if (eq (g1, 1)) return 1;
  if (eq (g2, 0)) return 1;
  if (eq (g2, 1)) return g1;
  if (eq (g2, apply (MMX_OVER, generic (1), generic (2))))
    return generic ("sqrt") (g1);
  if (frac_flag) {
    generic abs; int sgn;
    signed_decompose (g2, abs, sgn);
    if (sgn < 0) return apply (MMX_OVER, 1, simple_pow (g1, abs));
  }
  return apply (MMX_POWER, g1, g2);
}

/******************************************************************************
* Interface
******************************************************************************/

generic
simple_construct (const rational& x) {
  if (den (x) == 1) return generic (as_string (num (x)));
  return generic (as_string (num (x))) / generic (as_string (den (x)));
}

void
define_simple () {
  define_constant<always,bool> ("simple_evaluator?", true);
  define_constructor<always> (simple_construct);
  define<always> (MMX_MINUS, simple_neg);
  define<always> (MMX_PLUS , simple_add);
  define<always> (MMX_MINUS, simple_sub);
  define<always> (MMX_TIMES, simple_mul);
  define<always> (MMX_OVER , simple_div);
  define<always> (MMX_POWER, simple_pow);
}

bool
using_simple () {
  return current_ev->contains ("simple_evaluator?");
}

evaluator
simple_evaluator () {
  evaluator ev= base_evaluator ();
  select_evaluator (ev);
  define_prerequisites ();
  define_simple ();
  restore_evaluator ();
  return ev;
}

} // namespace mmx
