
/******************************************************************************
* MODULE     : mmx-light.cpp
* DESCRIPTION: Light command line evaluator of expressions
* COPYRIGHT  : (C) 2006  Joris van der Hoeven and Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <unistd.h>
#include "tuple.hpp"
#include "list.hpp"
#include "glue.hpp"
#include "mmx_syntax.hpp"
#include "shell.hpp"

using namespace mmx;

/*
 * Glue new functions for the interface
 */

static bool
toggle_quiet_mode (void) {
  quiet_mode = !quiet_mode;
  return quiet_mode;
}

static bool
toggle_debug_mode (void) {
  debug_mode = !debug_mode;
  return debug_mode;
}

static bool
toggle_type_mode (void) {
  type_mode = !type_mode;
  return type_mode;
}

static bool
toggle_time_mode (void) {
  time_mode = !time_mode;
  return time_mode;
}

static bool
toggle_math_mode (void) {
  math_mode = !math_mode;
  return math_mode;
}

static generic
mmx_quit () {
  cerr.flush ();
  cout.flush ();
  restore_evaluator ();
  exit (0);
}

static void
define_interface (void) {
  define<always> ("quiet_mode", toggle_quiet_mode);
  define<always> ("debug_mode", toggle_debug_mode);
  define<always> ("type_mode", toggle_type_mode);
  define<always> ("time_mode", toggle_time_mode);
  define<always> ("math_mode", toggle_math_mode);
  define<always> ("quit", mmx_quit);
}

static void
handle_evaluation (const mmx::string& s) {
  nat start, total;
  list<syntax_error> errors;
  generic in, out;
  static generic nil_generic = generic();

  if (N(s)==0) return;
  start= mmx_user_time ();
  in = mmx_parse ("", input_number (), s, errors);
  
  if (shell_terminal_print_syntax_errors (errors)) return;
  
  if (eq (in, nil_generic)) return;

  out= current_ev->eval (in);;
  total= mmx_user_time () - start;

  if (!is<tuple<generic> > (out) || N (as<tuple<generic> > (out)) > 1) {
    if (is<exception> (out))
      shell_terminal_print_exception (out);
    else {
      generic t= "None";
      if (type_mode) t= type_name (type (out));
      if (texmacs_mode)
	shell_texmacs_output (out, t);
      else
	shell_terminal_output (out, t);
    }
  }
  if (time_mode && total >= 1) cout << "Computed in " << total << " ms\n";
  cout.flush();
};

int
main (int argc, char** argv) {
  mmx::string in;
  list<mmx::string> to_load;
  bool neof;

  select_evaluator (standard_evaluator ());
  define_interface ();
  to_load = shell_initialize (argc, argv);

  if (!isatty(STDIN_FILENO) && !texmacs_mode) {
    // run in batch mode
    shell_terminal_batch_input (in);
    handle_evaluation (in);
    mmx_quit();
  }

  // enter in interactive mode
  /* Rather put in the manual
#ifdef HAVE_LIBREADLINE
  if (!texmacs_mode)
    cout << "// Type [Meta] + [Return] to insert a newline.\n";
#endif
  */
  while (true) {
    if (texmacs_mode)
      neof = shell_texmacs_input (in);
    else
      neof = shell_terminal_input (in);
    if (in == mmx::string ("quit\n") || in == mmx::string ("exit\n")
	|| in == mmx::string ("bye\n"))
      mmx_quit();
    handle_evaluation (in);
    if (!neof) {
      cout << "\n";
      mmx_quit();
    }
  }
}
