
/******************************************************************************
* MODULE     : environment.cpp
* DESCRIPTION: Environments for the base evaluator
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "environment.hpp"
#include "literal.hpp"
#include "routine.hpp"
#include "alias.hpp"
namespace mmx {

/******************************************************************************
* Environment access
******************************************************************************/

generic
environment_rep::name () const {
  array<generic> vars (entries (bindings));
  if (nil (next)) return gen ("global", vars);
  else return gen ("local", append (vars, array<generic> (next->name ())));
}

bool
environment_rep::contains (const generic& var) const {
  if (bindings->contains (var)) return true;
  else if (nil (next)) return false;
  else if (next->contains (var)) {
    environment_rep* me= const_cast<environment_rep*> (this);
    me->bindings (var)= next[var];
    return true;
  }
  else return false;
}

bool
environment_rep::read (const generic& var, generic& val) const {
  if (bindings->read (var, val)) return true;
  else if (nil (next)) return false;
  else if (next->read (var, val)) {
    environment_rep* me= const_cast<environment_rep*> (this);
    me->bindings (var)= val;
    return true;
  }
  else return false;
}

generic
environment_rep::get (const generic& var) const {
  generic val;
  if (bindings->read (var, val)) return val;
  else if (nil (next)) return val;
  else if (next->read (var, val)) {
    environment_rep* me= const_cast<environment_rep*> (this);
    me->bindings (var)= val;
    return val;
  }
  else return val;
}

list<string>
environment_rep::strings_for_completion () const {
  list<string> vars;
  iterator<generic> tmp = entries (bindings);
  for ( ; busy (tmp); ++tmp)
    if (is<literal> (*tmp)) 
      vars = cons (*as<literal>(*tmp), vars);
  return vars;
}

/******************************************************************************
* Converters
******************************************************************************/

inline generic tag (const generic& name, nat id) {
  return gen (name, as<generic> (id)); }
inline generic tag (const generic& name, nat id1, nat id2) {
  return gen (name, as<generic> (id1), as<generic> (id2)); }

void
environment_rep::ensure_up_to_date () const {
  if (!nil (next)) {
    next->ensure_up_to_date ();
    if (next_serial != next->serial) {
      environment_rep* me= const_cast<environment_rep*> (this);
      me->serial++;
      me->next_serial= next->serial;
    }
  }
}

void
environment_rep::set_converter (nat src, nat dest, const generic& val, nat trv)
{
  // add the converter
  set (tag ("converter" , src, dest), val);
  set (tag ("transitive", src, dest), as<generic> (trv));

  // add the converter to the list of converters
  generic t= tag ("converters", src);
  if (!contains (t)) set (t, as<generic> (array<nat> ()));
  set (t, as<generic> (cons<nat> (dest, as<array<nat> > (get (t)))));

  // time stamp
  environment_rep* me= const_cast<environment_rep*> (this);
  me->serial++;
}

generic
environment_rep::get_converter (nat src, nat dest) const {
  //cout << "Convert: " << src << ", " << dest << "\n";
  if (src == dest)
    return as<generic> (identity_routine (array<nat> (dest, src)));
  if (dest == 0) {
    if (is_alias_type (src)) {
      routine r;
      alias_reader (src, r);
      return as<generic> (r);
    }
    else return as<generic> (identity_routine (array<nat> (dest, src)));
  }

  ensure_up_to_date ();
  generic stamp;
  bool ok= read (tag ("cache/converters", src), stamp);
  if (!ok || !is<nat> (stamp) || serial != as<nat> (stamp)) {
    environment_rep* me= const_cast<environment_rep*> (this);
    table<routine,nat> funt;
    table<nat,nat> trvt;
    array<nat> todo;

    // setup identity converter
    routine id= identity_routine (array<nat> (src, src));
    me->set (tag ("cache/converter", src, src), as<generic> (id));
    funt (src)= id;
    trvt (src)= 3;
    todo << src;

    // transitive closure
    for (nat i=0; i<N(todo); i++) {
      nat cur= todo[i];
      generic cvs;
      if (!read (tag ("converters", cur), cvs)) continue;
      generic all= gen ("converters", as<generic> (cur));
      array<nat> succ= as<array<nat> > (cvs);
      for (nat j=0; j<N(succ); j++) {
	if (!funt->contains (succ[j])) {
	  // FIXME: also test for improvement of transitivity
	  nat     next    = succ[j];
	  generic next_fun= get (tag ("converter", cur, next));
	  generic next_trv= get (tag ("transitive", cur, next));
	  assert (is<routine> (next_fun) && is<nat> (next_trv));
	  nat trv1= trvt[cur], trv2= as<nat> (next_trv);
	  if ((trv1&2) != 0 || (trv2&1) != 0) {
	    routine fun= compose (as<routine> (next_fun), funt[cur]);
	    nat     trv= (trv1&1) + (trv2&2);
	    me->set (tag ("cache/converter", src, next), as<generic> (fun));
	    funt (next)= fun;
	    trvt (next)= trv;
	    todo << next;
	  }
	}
      }
    }

    // time stamp
    me->set (tag ("cache/converters", src), as<generic> (serial));
  }

  generic fun;
  if (read (tag ("cache/converter", src, dest), fun)) return fun;
  return as<generic> (routine ());
}

} // namespace mmx
