
/******************************************************************************
* MODULE     : generic.cpp
* DESCRIPTION: Generic expressions
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "generic_internal.hpp"
#include "evaluator.hpp"
#include "rational.hpp"
#include "array.hpp"
#include "mmx_syntax.hpp"
#include "memoize.hpp"
namespace mmx {

/******************************************************************************
* Automatic addition of new types
******************************************************************************/

nat
new_type_id () {
  static nat counter= 2;
  assert (counter < 65000);
  return counter++;
}

nat type_information<generic>::id= 0;
// type identifier 1 for unknown type

/******************************************************************************
* Printing generic expressions
******************************************************************************/

generic
melt (const generic& x1) {
  return x1.rep->melt_me ();
}

generic
fuse (const generic& x1) {
  MEMOIZE_UNARY (eq_memoizer,generic,generic,fuse,x1,
		 x1.rep->fuse_me ());
}

generic
flatten (const generic& g) {
  MEMOIZE_UNARY (std_memoizer,generic,generic,flatten,g,
		 g.rep->expression());
}

ostream&
operator << (ostream& out, const generic& g) {
  select_output_evaluator ();
  generic f= flatten (g);
  //lisp_print (out, f);
  mmx_print (out, f);
  restore_evaluator ();
  return out;
}

/******************************************************************************
* N-ary operations
******************************************************************************/

void
xapply_sub (array<generic>& v, const generic& g) {
  if (is<compound> (g) && N(g) >= 1 && eq (g[0], MMX_COMMA)) {
    int i, n= N(g);
    for (i=1; i<n; i++)
      xapply_sub (v, g[i]);
  }
  else v << g;
}

generic
xapply (const generic& f, const generic& args) {
  array<generic> v;
  v << f;
  xapply_sub (v, args);
  return array_to_compound (v);
}

generic comma () {
  return gen (generic (MMX_COMMA)); }
generic comma (const generic& g1, const generic& g2) {
  return gen (MMX_COMMA, g1, g2); }
generic xaccess (const generic& f, const generic& args) {
  return xapply (MMX_ACCESS, comma (f, args)); }
generic xtuple (const generic& g) {
  return xapply (MMX_TUPLE, g); }
generic xsqtuple (const generic& g) {
  return xapply (MMX_SQTUPLE, g); }
generic xrow (const generic& g) {
  return xapply (MMX_ROW, g); }
generic apply (const generic& g, const generic& x) {
  return gen (g, x); }
generic apply (const generic& g, const generic& x, const generic& y) {
  return gen (g, x, y); }
generic apply (const generic& g, const generic& x,
	       const generic& y, const generic& z) {
  return gen (g, x, y, z); }
generic apply (const generic& g, const array<generic>& args) {
  return array_to_compound (append (array<generic> (g), args)); }
generic access (const generic& g, const generic& x) {
  return gen (MMX_ACCESS, g, x); }
generic access (const generic& g, const generic& x, const generic& y) {
  return gen (MMX_ACCESS, g, x, y); }
generic access (const generic& g, const array<generic>& args) {
  array<generic> v= array<generic> (generic (MMX_ACCESS), g);
  return array_to_compound (append (v, args)); }

/******************************************************************************
* Predicates
******************************************************************************/

bool
is_scalar (const generic& g) {
  generic r= current_ev->apply ("scalar?", g);
  assert (is<bool> (r));
  return as<bool> (r);
}

bool is_central (const generic& g) { (void) g; return true; }
bool is_operator (const generic& g) { (void) g; return false; }

bool scalar_expand () { return true; }
bool scalar_power_expand () { return true; }
bool scalar_power_multiply () { return true; }
bool central_expand () { return false; }
bool central_power_expand () { return true; }
bool central_power_multiply () { return true; }

/******************************************************************************
* Evaluation
******************************************************************************/

generic
eval (const generic& x) {
  return current_ev->eval (x);
}

/******************************************************************************
* Constructors
******************************************************************************/

generic::generic (int i):
  rep (new generic_concrete_rep<rational> (i))
{
  *this= current_ev->construct (*this);
}

generic::generic (nat i):
  rep (new generic_concrete_rep<rational> (i))
{
  *this= current_ev->construct (*this);
}

generic::generic (const integer& i):
  rep (new generic_concrete_rep<rational> (i))
{
  *this= current_ev->construct (*this);
}

generic::generic (const rational& x):
  rep (new generic_concrete_rep<rational> (x))
{
  *this= current_ev->construct (*this);
}

/******************************************************************************
* Constants
******************************************************************************/

generic constant_helper<generic>::pi () { return MMX_PI; }
generic constant_helper<generic>::log2 () { return log (generic (2)); }
generic constant_helper<generic>::euler () { return MMX_EULER; }
generic constant_helper<generic>::catalan () { return MMX_CATALAN; }
generic constant_helper<generic>::imaginary () { return MMX_I; }
generic constant_helper<generic>::nan () { return MMX_NAN; }
generic constant_helper<generic>::fuzz () { return MMX_FUZZ; }
generic constant_helper<generic>::infinity () { return MMX_INFINITY; }
generic constant_helper<generic>::plus_infinity () { return MMX_INFINITY; }
generic constant_helper<generic>::minus_infinity () { return -infinity (); }

/******************************************************************************
* Basic arithmetic
******************************************************************************/

generic operator - (const generic& x1) {
  return current_ev->apply (MMX_MINUS, x1); }
generic operator + (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_PLUS, x1, x2); }
generic operator - (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_MINUS, x1, x2); }
generic operator * (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_TIMES, x1, x2); }
generic operator / (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_OVER, x1, x2); }
generic quo (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_DIV, x1, x2); }
generic rem (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_MOD, x1, x2); }
generic big_add (const array<generic>& v) {
  return current_ev->xapply (apply (MMX_PLUS, v)); }
generic big_mul (const array<generic>& v) {
  return current_ev->xapply (apply (MMX_TIMES, v)); }

/******************************************************************************
* Elementary functions
******************************************************************************/

generic pow (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_POWER, x1, x2); }
generic pow (const generic& x1, const int& x2) {
  return current_ev->apply (MMX_POWER, x1, generic (x2)); }
generic pow (const generic& x1, const nat& x2) {
  return current_ev->apply (MMX_POWER, x1, generic (x2)); }
generic pow (const int& x1, const generic& x2) {
  return current_ev->apply (MMX_POWER, generic (x1), x2); }
generic pow (const nat& x1, const generic& x2) {
  return current_ev->apply (MMX_POWER, generic (x1), x2); }
template<> generic sqrt (const generic& x1) {
  return current_ev->apply (MMX_SQRT, x1); }
template<> generic exp (const generic& x1) {
  return current_ev->apply (MMX_EXP, x1); }
template<> generic log (const generic& x1) {
  return current_ev->apply (MMX_LOG, x1); }
template<> generic cos (const generic& x1) {
  return current_ev->apply (MMX_COS, x1); }
template<> generic sin (const generic& x1) {
  return current_ev->apply (MMX_SIN, x1); }
template<> generic tan (const generic& x1) {
  return current_ev->apply (MMX_TAN, x1); }
generic acos (const generic& x1) {
  return current_ev->apply (MMX_ARCCOS, x1); }
generic asin (const generic& x1) {
  return current_ev->apply (MMX_ARCSIN, x1); }
generic atan (const generic& x1) {
  return current_ev->apply (MMX_ARCTAN, x1); }
generic trig (const generic& x1) {
  return xsqtuple (comma (cos (x1), sin (x1))); }

/******************************************************************************
* Predicates
******************************************************************************/

bool
operator == (const generic& x1, const generic& x2) {
  if (type (x1) == type (x2)) return x1.rep->is_equal (x2);
  generic r= current_ev->apply (MMX_EQUAL, x1, x2);
  if (!is<bool> (r)) return false;
  return as<bool> (r);
}

bool
operator != (const generic& x1, const generic& x2) {
  if (type (x1) == type (x2)) return !x1.rep->is_equal (x2);
  generic r= current_ev->apply (MMX_UNEQUAL, x1, x2);
  if (!is<bool> (r)) return true;
  return as<bool> (r);
}

bool
operator < (const generic& x1, const generic& x2) {
  generic r= current_ev->apply (MMX_LESS, x1, x2);
  assert (is<bool> (r));
  return as<bool> (r);
}

bool
operator <= (const generic& x1, const generic& x2) {
  generic r= current_ev->apply (MMX_LESSEQ, x1, x2);
  assert (is<bool> (r));
  return as<bool> (r);
}

bool
operator > (const generic& x1, const generic& x2) {
  generic r= current_ev->apply (MMX_GTR, x1, x2);
  assert (is<bool> (r));
  return as<bool> (r);
}

bool
operator >= (const generic& x1, const generic& x2) {
  generic r= current_ev->apply (MMX_GTREQ, x1, x2);
  assert (is<bool> (r));
  return as<bool> (r);
}

/******************************************************************************
* Real and complex structures
******************************************************************************/

generic min (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_MIN, x1, x2); }
generic max (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_MAX, x1, x2); }
generic abs (const generic& x1) {
  return current_ev->apply (MMX_ABS, x1); }
generic arg (const generic& x1) {
  return current_ev->apply (MMX_ARG, x1); }
generic Re (const generic& x1) {
  return current_ev->apply (MMX_RE, x1); }
generic Im (const generic& x1) {
  return current_ev->apply (MMX_IM, x1); }
generic conj (const generic& x1) {
  return current_ev->apply (MMX_CONJ, x1); }
generic gaussian (const generic& x1, const generic& x2) {
  return x1 + x2 * Imaginary (generic); }
generic polar (const generic& x1, const generic& x2) {
  return x1 * exp (x2 * Imaginary (generic)); }

/******************************************************************************
* Calculus
******************************************************************************/

generic derive (const generic& x1) {
  return current_ev->apply (MMX_DERIVE, x1); }
generic derive (const generic& x1, const generic& v) {
  MEMOIZE_BINARY (std_memoizer,generic,generic,generic,derive,x1,v,
		  current_ev->apply (MMX_DERIVE, x1, v)); }
generic prime (const generic& x1) {
  return current_ev->apply ("'", x1); }
generic integrate (const generic& x1) {
  return current_ev->apply (MMX_INTEGRATE, x1); }
generic integrate (const generic& x1, const generic& v) {
  return current_ev->apply (MMX_INTEGRATE, x1, v); }
generic substitute (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_SUBSTITUTE, x1, x2); }
generic compose (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_COMPOSE, x1, x2); }
generic dilate (const generic& x1, const generic& x2) {
  return current_ev->apply ("dilate", x1, x2); }
generic solve (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_SOLVE, x1, x2); }

/******************************************************************************
* Analytic continuation
******************************************************************************/

generic sqrt_init (const generic& x1, const generic& x2) {
  return current_ev->apply ("sqrt_init", x1, x2); }
generic log_init (const generic& x1, const generic& x2) {
  return current_ev->apply ("log_init", x1, x2); }
generic acos_init (const generic& x1, const generic& x2) {
  return current_ev->apply ("acos_init", x1, x2); }
generic asin_init (const generic& x1, const generic& x2) {
  return current_ev->apply ("asin_init", x1, x2); }
generic atan_init (const generic& x1, const generic& x2) {
  return current_ev->apply ("atan_init", x1, x2); }
generic integrate_init (const generic& x1, const generic& x2) {
  return current_ev->apply ("integrate_init", x1, x2); }
generic solve_lde_init (const generic& x1, const generic& x2) {
  return current_ev->apply ("solve_lde_init", x1, x2); }

/******************************************************************************
* Miscellaneous
******************************************************************************/

generic operator << (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_LESSLESS, x1, x2); }
generic operator >> (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_GTRGTR, x1, x2); }
generic shift_exp (const generic& x1, const generic& x2) {
  return current_ev->apply ("shift_exp", x1, x2); }
generic operator | (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_OR, x1, x2); }
generic operator & (const generic& x1, const generic& x2) {
  return current_ev->apply (MMX_AND, x1, x2); }
generic lres (const generic& x1, const generic& x2) {
  return current_ev->apply ("lres", x1, x2); }
generic rres (const generic& x1, const generic& x2) {
  return current_ev->apply ("rres", x1, x2); }
generic uniform (const generic& x1, const generic& x2) {
  return current_ev->apply ("uniform", x1, x2); }
generic specialize (const generic& x1, const generic& x2) {
  return current_ev->apply ("specialize", x1, x2); }

} // namespace mmx
