
/******************************************************************************
* MODULE     : integer.hpp
* DESCRIPTION: Integers
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_INTEGER_HPP
#define __MMX_INTEGER_HPP
#include "mmx_gmp.hpp"
namespace mmx {

class string;
char* as_charp (const string& s);
void free_charp (char* s);

class integer;
class integer_rep;

class integer_rep REP_STRUCT {
private:
  mpz_t x;
public:
  inline integer_rep () { mpz_init (x); }
  inline ~integer_rep () { mpz_clear (x); }
  friend class integer;
  friend class rational;
};

class integer {
INDIRECT_PROTO (integer, integer_rep)
private:
  inline mpz_t& operator * () const { return rep->x; }

public:
  // constructors
  inline integer (): rep (new integer_rep ()) {}
  inline integer (signed int i):
    rep (new integer_rep ()) { mpz_set_si (rep->x, i); }
  inline integer (unsigned int i):
    rep (new integer_rep ()) { mpz_set_ui (rep->x, i); }
  inline integer (signed short int i):
    rep (new integer_rep ()) { mpz_set_si (rep->x, i); }
  inline integer (unsigned short int i):
    rep (new integer_rep ()) { mpz_set_ui (rep->x, i); }
  inline integer (signed long int i):
    rep (new integer_rep ()) { mpz_set_si (rep->x, i); }
  inline integer (unsigned long int i):
    rep (new integer_rep ()) { mpz_set_ui (rep->x, i); }
  inline integer (char* s, int base= 0):
    rep (new integer_rep ()) { mpz_set_str (rep->x, s, base); }
  inline integer (const string& s, int base= 0):
    rep (new integer_rep ()) {
      char* r= as_charp (s);
      mpz_set_str (rep->x, r, base);
      free_charp (r); }

  // conversion
  inline friend int as_int (const integer& x) {
    return (int) mpz_get_si (*x); }

  // input/output
  friend string as_string (const integer& x);

  // basic arithmetic
  inline friend integer operator - (const integer& x1) {
    integer r; mpz_neg (*r, *x1); return r; }
  inline friend integer operator + (const integer& x1, const integer& x2) {
    integer r; mpz_add (*r, *x1, *x2); return r; }
  inline friend integer operator - (const integer& x1, const integer& x2) {
    integer r; mpz_sub (*r, *x1, *x2); return r; }
  inline friend integer operator * (const integer& x1, const integer& x2) {
    integer r; mpz_mul (*r, *x1, *x2); return r; }
  inline friend integer
  operator / (const integer& x1, const integer& x2) PARTIAL {
    integer r; mpz_fdiv_r (*r, *x1, *x2);
    domain_check (mpz_sgn (*r) == 0); // exact division?
    mpz_fdiv_q (*r, *x1, *x2); return r; }
  inline friend integer quo (const integer& x1, const integer& x2) {
    integer r; mpz_fdiv_q (*r, *x1, *x2); return r; }
  inline friend integer rem (const integer& x1, const integer& x2) {
    integer r; mpz_fdiv_r (*r, *x1, *x2); return r; }
  inline friend integer operator << (const integer& x1, long int shift) {
    integer r; mpz_mul_2si (*r, *x1, shift); return r; }
  inline friend integer operator >> (const integer& x1, long int shift) {
    integer r; mpz_mul_2si (*r, *x1, -shift); return r; }
  inline friend integer& operator <<= (integer& x1, long int shift) {
    return x1= x1 << shift; }
  inline friend integer& operator >>= (integer& x1, long int shift) {
    return x1= x1 >> shift; }
  inline friend integer sqrt (const integer& x1) {
    integer r; mpz_sqrt (*r, *x1); return r; }
  inline friend integer pow (const integer& x1, const int& i) {
    assert (i >= 0);
    integer r; mpz_pow_ui (*r, *x1, (nat) i); return r; }
  inline friend integer pow (const integer& x1, const nat& i) {
    integer r; mpz_pow_ui (*r, *x1, i); return r; }
  inline friend integer abs (const integer& x1) {
    integer r; mpz_abs (*r, *x1); return r; }

  // comparing integer numbers
  inline friend bool operator == (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) == 0; }
  inline friend bool operator != (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) != 0; }
  inline friend bool operator < (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) < 0; }
  inline friend bool operator > (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) > 0; }
  inline friend bool operator <= (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) <= 0; }
  inline friend bool operator >= (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) >= 0; }
  inline friend int sign (const integer& x1) {
    int r= mpz_sgn (*x1); return (r<0? -1: (r==0? 0: 1)); }
  inline friend integer min (const integer& x1, const integer& x2) {
    return x1 <= x2? x1: x2; }
  inline friend integer max (const integer& x1, const integer& x2) {
    return x1 >= x2? x1: x2; }

  // miscellaneous functions
  inline friend nat hash (const integer& x) {
    __mpz_struct &rep= (*x)[0];
    if (rep._mp_size == 0) return 0;
    return ((nat) (rep._mp_d[0])) ^ ((nat) (rep._mp_size << 11));
  }
  inline friend nat bit_size (const integer& x) {
    return mpz_sizeinbase (*x, 2); }
  inline friend nat limb_size (const integer& x) {
    return mpz_size (*x); }
  
  //gcd of two integers
  inline friend integer gcd (const integer& a, const integer& b) {
    integer r; mpz_gcd(*r, *a, *b ); return r; }
  
  friend class rational;
  friend integer num (const rational& x1);
  friend integer den (const rational& x1);
  template<typename V> friend class floating;
  template<typename V> friend integer as_integer (const floating<V>& x);
};
INDIRECT_IMPL (integer, integer_rep)

PRINT_SUGAR(,integer);
ARITH_INT_SUGAR(,integer,integer);

STMPL inline bool eq (const integer& x, const integer& y) {
  return x == y; }
STMPL inline bool neq (const integer& x, const integer& y) {
  return x != y; }

}
#endif // __MMX_INTEGER_HPP
