
/******************************************************************************
* MODULE     : mmx_lexer.cpp
* DESCRIPTION: Lexical analysis of mathemagix files
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* NOTE: This file is included from mmx_parser.ypp
*       It is distributed apart together with the sources
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "mmx_syntax.hpp"
#include "parse_tools.hpp"
namespace mmx {

#define s     (data->lex_string)
#define n     (data->lex_length)

#define start (data->lex_start)
#define prev  (data->lex_prev)
#define pos   (data->lex_pos)

#define start_line (data->lex_start_line)
#define prev_line  (data->lex_prev_line)
#define line       (data->lex_line)

#define start_begin_line (data->lex_start_begin_line)
#define prev_begin_line  (data->lex_prev_begin_line)
#define begin_line       (data->lex_begin_line)

#define file  (data->lex_file_name)
#define input (data->lex_input_number)

#define INC_POS   { pos++; }
#define ADD_POS(z){ pos += z; }
#define INC_LINE  { pos++; line++; begin_line = pos; }
#define SET_PREV  { prev = pos; prev_line = line; prev_begin_line = begin_line; }
#define SAVE_START { start = pos; start_line = line; start_begin_line = begin_line; }
#define RESTORE_START { pos = start; line = start_line; begin_line = start_begin_line; }

#define produce(code) { \
  *lval = string (s + start, pos - start); \
  source_insert (*lval, source_location (*lval, file, input,	\
		source_position(start, start_line, start - start_begin_line),\
	        source_position(pos, line, pos - begin_line))); \
  return code; }

#define test(c,code) \
  if ((pos<n) && (s[pos]==c)) { INC_POS; produce (code); }

#define keyword(c,k,code) \
  if ((s[pos]==c) && test_keyword(k,data)) produce (code);

#define is_alpha(c) \
  ((((c)>='0') && ((c)<='9')) || \
   (((c)>='A') && ((c)<='Z')) || \
   (((c)>='a') && ((c)<='z')) || \
   (((c)=='_') || ((c)=='?')))

static bool
test_keyword (char* k, parse_data* data) {
  RESTORE_START;
  while ((*k)!='\0') {
    if (pos>=n) return false;
    if (s[pos]!=(*k)) return false;
    INC_POS;
    k++;
  }
  if (pos>=n) return true;
  if (is_alpha (s[pos])) return false;
  return true;
}

int
mmx_lex (generic *lval, parse_data* data) {
  if (pos == n) {
    s= NULL;
    *lval= generic ();
    return 0;
  }

  SET_PREV;
  char c= s[pos];
  while ((c<=' ') || (c>'~')) {
    if ((c == '\n') || (c == '\r')) 
      INC_LINE
    else
      INC_POS;
    if (pos == n) {
      s= NULL;
      *lval= generic ();
      return 0;
    }
    c= s[pos];
  }

  SAVE_START;
  INC_POS;
  switch (c) {
  case '!':
    test ('=', NOT_EQUAL);
    if ((pos<n) && (s[pos]=='<')) {
      INC_POS;
      test ('=', NOT_LEQ);
      produce (NOT_LESS);
    }
    if ((pos<n) && (s[pos]=='>')) {
      INC_POS;
      test ('=', NOT_GEQ);
      produce (NOT_GREATER);
    }
    produce (NOT);
  case '\042':
    while (pos<n) {
      if ((s[pos]=='\\') && ((pos+2)<n)) { ADD_POS(2); continue; }
      if (s[pos]=='\042') { INC_POS; break; }
      INC_POS;
    }
    produce (STRING);
  case '#':
    produce (SIZE);
  case '$':
  case '%':
  case '&':
    produce (c);
  case '\047':
    produce (PRIME);
  case '(':
  case ')':
    produce (c);
  case '*':
    test ('=', TIMES_ASSIGN);
    produce (TIMES);
  case '+':
    test ('=', PLUS_ASSIGN);
    test ('+', INC);
    produce (PLUS);
  case ',':
    produce (c);
  case '-':
    test ('=', MINUS_ASSIGN);
    test ('-', DEC);
    test ('>', INTO);
    produce (MINUS);
  case '.':
    test ('.', RANGEEQ);
    produce (ACCESS);
  case '/':
    test ('\\', AND);
    if (s[pos]=='/') {
      INC_POS;
      while ((pos<n) && (s[pos]!='\n')) {
	INC_POS;
      }
      if (pos<n) { INC_LINE; }
      return mmx_lex (lval, data);
    }
    if (s[pos]=='{') {
      nat level= 1;
      INC_POS;
      while ((pos+1<n) && (level>0)) {
	if ((s[pos]=='/') && (s[pos+1]=='{')) { level++; ADD_POS(2); continue; }
	if ((s[pos]=='}') && (s[pos+1]=='/')) { level--; ADD_POS(2); continue; }
	ADD_POS(2);
      }
      if (level>0) pos=n;
      return mmx_lex (lval, data);
    }
    if (s[pos] == '\"') {
      INC_POS;
      while ((pos+1<n) && ((s[pos]!='\"') || (s[pos+1]!='/'))) {
	INC_POS;
      }
      if (pos+1 < n) {
	ADD_POS(2);
      }
      produce (STRING);
    }
    test ('=', OVER_ASSIGN);
    produce (OVER);
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    while ((pos<n) && (s[pos]>='0') && (s[pos]<='9')) INC_POS;
    if ((pos<n+1) && (s[pos]=='.') && (s[pos+1]>='0') && (s[pos+1]<='9')) {
      ADD_POS(2);
      while ((pos<n) && (s[pos]>='0') && (s[pos]<='9')) INC_POS;
      if ((pos<n+1) && ((s[pos]=='e') || (s[pos]=='E'))) {
	nat p= ((pos<n+2) && (s[pos+1]=='-') ? 2: 1);
	if ((s[pos+p]>='0') && (s[pos+p]<='9')) {
	  ADD_POS(p);
	  while ((pos<n) && (s[pos]>='0') && (s[pos]<='9')) INC_POS;
	}
      }
      produce (FLOATING);
    }
    else produce (INTEGER);
  case ':':
    test ('=', ASSIGN);
    test ('>', CAST);
    test (':', RANGE);
    if ((pos+1<n) && (s[pos]=='-') && (s[pos+1]=='>')) {
      ADD_POS(2); produce (MAPSTO);
    }
    produce (TYPE);
  case ';':
    produce (c);
  case '<':
    if ((pos<n) && (s[pos]=='<')) {
      if ((pos+1<n) && (s[pos+1]=='*')) {
	ADD_POS(2); produce (LEFT_FLUX_VAR); }
      if ((pos+1<n) && (s[pos+1]=='%')) {
	ADD_POS(2); produce (LEFT_FLUX_STR); }
      INC_POS; produce (LEFT_FLUX);
    }
    if ((pos+1<n) && (s[pos]=='=') && (s[pos+1]=='>')) {
      ADD_POS(2); produce (EQUIVALENT);
    }
    test ('=', LEQ);
    produce (LESS);
  case '=':
    test ('=', DEFINE);
    test ('>', IMPLIES);
    produce (EQUAL);
  case '>':
    test ('>', RIGHT_FLUX);
    test ('=', GEQ);
    produce (GREATER);
  case '?':
    produce (c);
  case '@':
    test ('+', OPLUS);
    test ('-', OMINUS);
    test ('*', OTIMES);
    test ('/', OOVER);
    produce (COMPOSE);
  case 'A':
  case 'B':
  case 'C':
  case 'D':
  case 'E':
  case 'F':
  case 'G':
  case 'H':
  case 'I':
  case 'J':
  case 'K':
  case 'L':
  case 'M':
  case 'N':
  case 'O':
  case 'P':
  case 'Q':
  case 'R':
  case 'S':
  case 'T':
  case 'U':
  case 'V':
  case 'W':
  case 'X':
  case 'Y':
  case 'Z':
    goto identifier;
  case '[':
    produce (c);
  case '\\':
    test ('/', OR);
    produce (c);
  case ']':
    produce (c);
  case '^':
    produce (POWER);
  case '_':
    goto identifier;
  case '`':
    produce (BACKPRIME);
  case 'a':
    if (pos<n) {
      keyword ('b', "abstract", ABSTRACT);
      keyword ('d', "add", ADD);
      keyword ('n', "and", SEQAND);
    }
    goto identifier;
  case 'b':
    if (pos<n) keyword ('r', "break", BREAK);
    goto identifier;
  case 'c':
    if ((pos+2)<n) {
      if (s[pos]=='a') {
	INC_POS;
	keyword ('s', "case", CASE);
	keyword ('t', "catch", CATCH);
      }
      if ((s[pos]=='o') && (s[pos+1]=='n')) {
	ADD_POS(2);
	keyword ('c', "concrete", CONCRETE);
	keyword ('s', "constant", CONSTANT);
	keyword ('t', "continue", CONTINUE);
      }
    }
    goto identifier;
  case 'd':
    if (pos<n) {
      if (s[pos]=='i') {
	INC_POS;
	keyword ('r', "direct", DIRECT);
	keyword ('v', "div", DIV);
      }
      keyword ('o', "do", DO);
    }
    goto identifier;
  case 'e':
    if ((pos+1)<n) {
      keyword ('l', "else", ELSE);
      keyword ('v', "evolutive", EVOLUTIVE);
      if (s[pos]=='x') {
	INC_POS;
	keyword ('i', "exists", EXISTS);
	keyword ('p', "export", EXPORT);
	if (((pos+2)<n) && (s[pos]=='t') && (s[pos+1]=='e')) {
	  ADD_POS(2);
	  keyword ('n', "extend", EXTEND);
	  keyword ('r', "extern", EXTERN);
	}
      }
    }
    goto identifier;
  case 'f':
    if (pos<n) {
      if ((pos+1<n) && (s[pos]=='o') && (s[pos+1]=='r')) {
	if ((pos+2<n) && (s[pos+2]=='a')) {
	  keyword ('o', "forall", FORALL);
	}
	else if ((pos+2<n) && (s[pos+2]=='e')) {
	  keyword ('o', "foreign", FOREIGN);
	}
	else {
	  keyword ('o', "for", FOR);
	}
      }
      keyword ('r', "from", FROM);
    }
    goto identifier;
  case 'g':
    if (pos<n) keyword ('e', "generate", GENERATE);
    goto identifier;
  case 'h':
    if (pos<n) {
      keyword ('a', "has", HAS);
      keyword ('i', "hidden", HIDDEN);
      keyword ('o', "holds", HOLDS);
    }
    goto identifier;
  case 'i':
    if (pos<n) {
      keyword ('f', "if", IF);
      keyword ('m', "import", IMPORT);
      if (s[pos]=='n') {
	INC_POS;
	if ((pos >= n) || (!is_alpha (s[pos]))) produce (IN);
	keyword ('d', "indirect", INDIRECT);
	// keyword ('f', "infix", INFIX);
	keyword ('h', "inherit", INHERIT);
	keyword ('l', "inline", INLINE);
	if ((pos+3<n) && (s[pos]=='t') && (s[pos+1]=='e') && (s[pos+2]=='r')) {
	  ADD_POS(3);
	  keyword ('a', "interactive", INTERACTIVE);
	  keyword ('n', "intern", INTERN);
	}
      }
    }
    goto identifier;
  case 'j':
    if (pos<n) keyword ('o', "join", JOIN);
    goto identifier;
  case 'k':
    goto identifier;
  case 'l':
    if (pos<n) {
      keyword ('a', "lambda", LAMBDA);
      keyword ('i', "literal", LITERAL);
      if (((pos+1)<n) && (s[pos]=='o')) {
	INC_POS;
	keyword ('c', "locked", LOCKED);
	keyword ('o', "loop", LOOP);
      }
    }
    goto identifier;
  case 'm':
    if (pos<n) {
      keyword ('e', "method", METHOD);
      keyword ('o', "mod", MOD);
    }
    goto identifier;
  case 'n':
    goto identifier;
  case 'o':
    if (pos<n) {
      keyword ('p', "operator", OPERATOR);
      keyword ('r', "or", SEQOR);
      keyword ('u', "outline", OUTLINE);
    }
    goto identifier;
  case 'p':
    if ((pos+1)<n) {
      keyword ('a', "packed", PACKED);
      keyword ('o', "postfix", POSTFIX);
      if (s[pos]=='r') {
        INC_POS;
	// keyword ('e', "prefix", PREFIX);
	keyword ('i', "private", PRIVATE);
	keyword ('o', "protected", PROTECTED);
      }
      keyword ('u', "public", PUBLIC);
    }
    goto identifier;
  case 'q':
    goto identifier;
  case 'r':
    if (pos<n) {
      keyword ('a', "raise", RAISE);
      keyword ('e', "return", RETURN);
    }
    goto identifier;
  case 's':
    if (pos<n) {
      keyword ('p', "split", SPLIT);
      keyword ('t', "step", STEP);
    }
    goto identifier;
  case 't':
    if (pos<n) {
      keyword ('h', "then", THEN);
      keyword ('r', "try", TRY);
    }
    goto identifier;
  case 'u':
    if (((pos+1)<n) && (s[pos]=='n')) {
      INC_POS;
      keyword ('p', "unpacked", UNPACKED);
      keyword ('t', "until", UNTIL);
    }
    goto identifier;
  case 'v':
    if (((pos+1)<n) && (s[pos]=='a')) {
      INC_POS;
      keyword ('l', "value", VALUE);
      keyword ('r', "variable", VARIABLE);
    }
    goto identifier;
  case 'w':
    if (pos<n) {
      keyword ('h', "while", WHILE);
      keyword ('i', "with", WITH);
    }
    goto identifier;
  case 'x':
    if (pos<n) keyword ('x', "xor", XOR);
    goto identifier;
  case 'y':
    if (pos<n) keyword ('i', "yield", YIELD);
    goto identifier;
  case 'z':
    goto identifier;
  case '{':
    produce (c);
  case '|':
    test ('|', VWHERE);
    produce (WHERE);
  case '}':
    produce (c);
  case '~':
    produce (TILDA);
  default:
    return mmx_lex (lval, data);

  identifier:
    while ((pos<n) && is_alpha (s[pos])) INC_POS;
    produce (IDENTIFIER);
  }
}

} // namespace mmx
